from snkmt.db.models.enums import Status, FileType


from snkmt.db.models.workflow import Workflow
from snkmt.db.models.rule import Rule
from snkmt.db.models.job import Job
from snkmt.db.models.file import File
from snkmt.db.models.error import Error


__all__ = [
    "Status",
    "FileType",
    "Workflow",
    "Rule",
    "Job",
    "File",
    "Error",
]



from pathlib import Path
from textual.app import App, ComposeResult
from textual.widgets import (
    Static,
    Footer,
    DataTable,
    Label,
    Collapsible,
    ListItem,
    ListView,
    Log,
)
from textual.screen import Screen, ModalScreen
from textual.containers import Container, Horizontal
from sqlalchemy.orm import Session
from sqlalchemy import select
from snkmt.db.models import Workflow, Rule
from snkmt.db.models.enums import Status, FileType
from rich.text import Text
from uuid import UUID
from snkmt.console.widgets import RuleTable, WorkflowTable, StyledProgress, StyledStatus
from snkmt.version import VERSION
from textual.reactive import reactive
from snkmt.db.models import Job


class AppHeader(Horizontal):
    """The header of the app."""

    def __init__(self, db_url: str, *args, **kwargs):
        self.db_url = db_url
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        yield Label(f"[b]snkmt[/] [dim]{VERSION}[/]", id="app-title")
        yield Label(f"Connected to: {self.db_url}", id="app-db-path")


class AppBody(Horizontal):
    """The body of the app"""


class LogFileModal(ModalScreen):
    """Modal to display log file text."""

    BINDINGS = [("escape", "app.pop_screen", "Pop screen")]

    def __init__(self, log_file: Path, *args, **kwargs):
        self.log_file = log_file
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        container = Container(id="modal-container")
        container.border_title = f"Logfile: {self.log_file}"
        container.border_subtitle = "Press esc to close."
        with container:
            yield Log(id="log-content", auto_scroll=False, highlight=True)

    def on_mount(self) -> None:
        """Load and display the log file content when modal is mounted."""
        log_widget = self.query_one(Log)

        try:
            if not self.log_file.exists():
                log_widget.write_line(f"Error: File '{self.log_file}' does not exist.")
                return

            if not self.log_file.is_file():
                log_widget.write_line(
                    f"Error: '{self.log_file}' is not a regular file."
                )
                return

            try:
                with open(self.log_file, "r", encoding="utf-8") as f:
                    content = f.read()

                if content.strip():
                    lines = content.splitlines()
                    log_widget.write_lines(lines)
                else:
                    log_widget.write_line("📄 File is empty.")

            except UnicodeDecodeError:
                log_widget.write_line(
                    "Error: Could not read file with any supported encoding."
                )
                log_widget.write_line(
                    "File may be binary or use an unsupported encoding."
                )

        except PermissionError:
            log_widget.write_line(f"Permission Error: Cannot read '{self.log_file}'.")
            log_widget.write_line(
                "You may not have sufficient permissions to access this file."
            )

        except FileNotFoundError:
            log_widget.write_line(
                f"File Not Found: '{self.log_file}' could not be found."
            )

        except OSError as e:
            log_widget.write_line(f"OS Error: {e}")
            log_widget.write_line("There was a system-level error reading the file.")

        except Exception as e:
            log_widget.write_line(f"Unexpected Error: {e}")
            log_widget.write_line(
                "An unexpected error occurred while reading the file."
            )


class WorkflowOverview(Container):
    workflow_id = reactive(None, recompose=True, always_update=True)

    def __init__(self, session: Session, *args, **kwargs) -> None:
        self.db_session = session
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        if self.workflow_id is None:
            yield Label("Please select a workflow to view details.")
        else:
            workflow = self.db_session.scalars(
                select(Workflow).where(Workflow.id == self.workflow_id)
            ).one_or_none()
            if not workflow:
                yield Label("Workflow not found. Something went wrong.")
            else:
                table = DataTable()
                table.add_column("Field", width=15)
                table.add_column("Value")
                table.cursor_type = "none"
                table.show_cursor = False
                table.show_header = False

                table.add_row(
                    Text("ID", justify="left", style="bold"),
                    Text(str(workflow.id), justify="left"),
                )
                table.add_row(
                    Text("Snakefile", justify="left", style="bold"),
                    Text(
                        workflow.snakefile or "N/A",
                        justify="left",
                        style="dim" if not workflow.snakefile else "",
                    ),
                )
                table.add_row(
                    Text("Started At", justify="left", style="bold"),
                    Text(
                        workflow.started_at.strftime("%Y-%m-%d %H:%M:%S")
                        if workflow.started_at
                        else "N/A",
                        justify="left",
                        style="dim" if not workflow.started_at else "",
                    ),
                )
                table.add_row(
                    Text("Updated At", justify="left", style="bold"),
                    Text(
                        workflow.updated_at.strftime("%Y-%m-%d %H:%M:%S")
                        if workflow.updated_at
                        else "N/A",
                        justify="left",
                        style="dim" if not workflow.updated_at else "",
                    ),
                )
                table.add_row(
                    Text("End Time", justify="left", style="bold"),
                    Text(
                        workflow.end_time.strftime("%Y-%m-%d %H:%M:%S")
                        if workflow.end_time
                        else "N/A",
                        justify="left",
                        style="dim" if not workflow.end_time else "",
                    ),
                )
                table.add_row(
                    Text("Status", justify="left", style="bold"),
                    StyledStatus(workflow.status),
                )
                table.add_row(
                    Text("Total Jobs", justify="left", style="bold"),
                    Text(str(workflow.total_job_count), justify="left"),
                )
                table.add_row(
                    Text("Jobs Finished", justify="left", style="bold"),
                    Text(str(workflow.jobs_finished), justify="left"),
                )
                table.add_row(
                    Text("Progress", justify="left", style="bold"),
                    StyledProgress(workflow.progress),
                )

                yield table


class WorkflowErrors(Container):
    workflow_id = reactive(None, recompose=True, always_update=True)

    def __init__(self, session: Session, *args, **kwargs) -> None:
        self.db_session = session
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        if self.workflow_id is None:
            yield Label("Please select a workflow to view errors.")
        else:
            workflow = self.db_session.scalars(
                select(Workflow).where(Workflow.id == self.workflow_id)
            ).one_or_none()
            if not workflow:
                yield Label("Workflow not found. Something went wrong.")
            elif not workflow.errors:
                yield Label("No errors. yay :)")
            else:
                failed_rules = self.db_session.scalars(
                    select(Rule)
                    .join(Job)
                    .where(Rule.workflow_id == workflow.id)
                    .where(Job.status == Status.ERROR)
                    .distinct()
                ).all()

                for rule in failed_rules:
                    # TODO rewrite this query to order by job time and probably limit it to some reasonalbe number?

                    failed_jobs = [
                        job for job in rule.jobs if job.status == Status.ERROR
                    ]

                    with Collapsible(
                        title=f"rule '{rule.name}' ({len(failed_jobs)} failed jobs)"
                    ):
                        labels = []
                        for job in failed_jobs:
                            log_files = [
                                f for f in job.files if f.file_type == FileType.LOG
                            ]
                            for lf in log_files:
                                if Path(lf.path).exists():
                                    labels.append(
                                        ListItem(
                                            Static(
                                                f"📄 Job {job.id}: {lf.path}",
                                                name=lf.path,
                                            )
                                        )
                                    )
                                else:
                                    self.log.debug(f"Log file {lf.path} doesn't exist.")

                        list_view = ListView(*labels)
                        list_view.styles.height = "auto"
                        yield list_view


class WorkflowDetail(Container):
    def __init__(self, session: Session, *args, **kwargs) -> None:
        self.db_session = session
        self.workflow_id = None
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        self.overview_section = WorkflowOverview(
            session=self.db_session, classes="subsection", id="workflow-overview"
        )
        self.rules_section = Container(classes="subsection", id="workflow-rules")
        self.errors_section = WorkflowErrors(
            session=self.db_session, classes="subsection", id="workflow-errors"
        )

        self.overview_section.border_title = "Workflow Info"
        self.rules_section.border_title = "Rules"
        self.errors_section.border_title = "Errors"

        yield self.overview_section
        yield self.rules_section
        yield self.errors_section

    def show_workflow(self, workflow_id: UUID) -> None:
        self.workflow_id = workflow_id

        # TODO make these self updating like ruletable.
        self.overview_section.workflow_id = workflow_id  # type: ignore
        self.errors_section.workflow_id = workflow_id  # type: ignore

        self.rules_section.remove_children()
        self.table = RuleTable(workflow_id, self.db_session)
        self.rules_section.mount(self.table)


class WorkflowDashboard(Container):
    BINDINGS = [
        ("tab", "focus_next", "Next"),
        ("shift+tab", "focus_previous", "Previous"),
    ]

    def __init__(self, session: Session) -> None:
        self.db_session = session
        self.table = WorkflowTable(session)
        super().__init__()

    def compose(self) -> ComposeResult:
        self.workflows = Container(classes="section", id="workflows")
        self.workflows.border_title = "Workflows"

        with self.workflows:
            yield self.table

        self.detail = WorkflowDetail(
            session=self.db_session, classes="section", id="detail"
        )
        self.detail.border_title = "Workflow Details"
        yield self.detail

    def on_data_table_row_selected(self, event: DataTable.RowSelected) -> None:
        """Handle row selection (clicking or pressing enter)."""
        if isinstance(event.data_table, WorkflowTable):
            workflow_id = UUID(event.row_key.value)
            self.log.debug(f"Selected workflow: {workflow_id}")
            self.detail.show_workflow(workflow_id)


class DashboardScreen(Screen):
    BINDINGS = [
        ("tab", "focus_next", "Next"),
        ("shift+tab", "focus_previous", "Previous"),
    ]

    def __init__(self, session: Session) -> None:
        super().__init__()
        self.session = session

    def compose(self) -> ComposeResult:
        yield AppHeader(str(self.session.bind.url))  # type: ignore
        yield WorkflowDashboard(self.session)
        yield Footer(id="footer")


class snkmtApp(App):
    """A Textual app for monitoring Snakemake workflows."""

    CSS_PATH = "snkmt.tcss"
    BINDINGS = [
        ("q", "quit", "Quit"),
        ("ctrl+c", "quit", "Quit"),
        ("tab", "focus_next", "Next"),
        ("shift+tab", "focus_previous", "Previous"),
    ]

    def __init__(self, db_session: Session):
        super().__init__()
        self.session = db_session

    def on_ready(self) -> None:
        self.title = "snkmt console"
        self.theme = "gruvbox"
        self.push_screen(DashboardScreen(self.session))

    def on_list_view_selected(self, event: ListView.Selected) -> None:
        list_item = event.item.children[0]
        if list_item.name:
            self.log.debug(f"log file selected: {list_item.name}")
            self.push_screen(LogFileModal(Path(list_item.name)))

    def action_focus_next(self) -> None:
        """Focus the next widget."""
        self.screen.focus_next()

    def action_focus_previous(self) -> None:
        """Focus the previous widget."""
        self.screen.focus_previous()


def run_app(db_session: Session):
    """Run the Textual app."""
    app = snkmtApp(db_session)
    app.run()



import uuid
from typing import TYPE_CHECKING, Optional

from sqlalchemy import ForeignKey, select, func, case
from sqlalchemy.orm import Mapped, mapped_column, relationship, Session
from datetime import datetime, timezone

from snkmt.db.models.base import Base

if TYPE_CHECKING:
    from snkmt.db.models.job import Job
    from snkmt.db.models.workflow import Workflow
    from snkmt.db.models.error import Error


class Rule(Base):
    __tablename__ = "rules"
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]
    workflow_id: Mapped[uuid.UUID] = mapped_column(ForeignKey("workflows.id"))
    updated_at: Mapped[datetime] = mapped_column(
        default=lambda: datetime.now(timezone.utc),
        onupdate=lambda: datetime.now(timezone.utc),
    )
    workflow: Mapped["Workflow"] = relationship("Workflow", back_populates="rules")
    total_job_count: Mapped[int] = mapped_column(default=0)  # from run info
    jobs_finished: Mapped[int] = mapped_column(default=0)
    jobs: Mapped[list["Job"]] = relationship(
        "Job", back_populates="rule", cascade="all, delete-orphan"
    )
    errors: Mapped[list["Error"]] = relationship(
        "Error", back_populates="rule", cascade="all, delete-orphan"
    )

    @property
    def progress(self) -> float:
        if self.total_job_count == 0:
            return 0.0
        return self.jobs_finished / self.total_job_count

    @classmethod
    def get_updated_since(
        cls,
        session: Session,
        workflow_id: uuid.UUID,
        timestamp,
        limit: Optional[int] = None,
    ):
        """Get rules for a workflow that have been updated since the given timestamp."""
        query = session.query(cls).filter(
            cls.workflow_id == workflow_id, cls.updated_at >= timestamp
        )
        if limit:
            query = query.limit(limit)
        return query.all()

    def get_job_counts(self, session):
        """Get all job counts in a single efficient query."""
        from snkmt.db.models.job import Job
        from snkmt.db.models.enums import Status

        result = session.execute(
            select(
                func.sum(case((Job.status == Status.RUNNING, 1), else_=0)).label(
                    "running"
                ),
                func.sum(case((Job.status == Status.ERROR, 1), else_=0)).label(
                    "failed"
                ),
                func.sum(case((Job.status == Status.SUCCESS, 1), else_=0)).label(
                    "success"
                ),
            ).where(Job.rule_id == self.id)
        ).one()

        running = result.running or 0
        failed = result.failed or 0
        success = result.success or 0

        # i dont think pending jobs are logged
        pending = self.total_job_count - running - failed - success

        return {
            "total": self.total_job_count,
            "running": running,
            "pending": pending,
            "failed": failed,
            "success": success,
        }



import shutil
import sys
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Tuple
from sqlalchemy import create_engine, inspect
from sqlalchemy.orm import sessionmaker, Session
from platformdirs import user_data_dir
from loguru import logger
from snkmt.db.models.version import (
    DB_VERSIONS,
    DB_MAX_VERSION,
    DB_MIN_VERSION,
    DBVersion,
    DBVersionError,
    null_db_version,
)
from snkmt.db.models.base import Base
from alembic.command import downgrade, upgrade
from alembic.config import Config as AlembicConfig


SNKMT_DIR = Path(user_data_dir(appname="snkmt", appauthor=False, ensure_exists=True))


logger.remove()
logger.add(sys.stderr)


class DatabaseNotFoundError(Exception):
    """Raised when the Snakemake DB file isn’t found and creation is disabled."""

    pass


class Database:
    """Simple connector for the Snakemake SQLite DB."""

    def __init__(
        self,
        db_path: Optional[str] = None,
        create_db: bool = True,
        auto_migrate: bool = True,
        ignore_version: bool = False,
    ):
        default_db_path = SNKMT_DIR / "snkmt.db"

        if db_path:
            db_file = Path(db_path)

        else:
            db_file = default_db_path

        if not db_file.parent.exists():
            if create_db:
                db_file.parent.mkdir(parents=True, exist_ok=True)
            else:
                raise DatabaseNotFoundError(f"No DB directory: {db_file.parent}")

        if not db_file.exists() and not create_db:
            raise DatabaseNotFoundError(f"DB file not found: {db_file}")

        self.db_path = str(db_file)
        self.engine = create_engine(
            f"sqlite:///{self.db_path}",
            echo=False,
            pool_size=10,
            max_overflow=20,
            pool_pre_ping=True,
            future=True,
        )
        self.SessionLocal = sessionmaker(
            autocommit=False, autoflush=True, bind=self.engine
        )
        self.session = self.get_session()
        self.auto_migrate = auto_migrate

        Base.metadata.create_all(bind=self.engine)

        current_version = self.get_version()
        latest_version = self.get_all_versions()[-1]

        if current_version == null_db_version:
            self.session.add(
                DBVersion(
                    id=latest_version.id,
                    major=latest_version.major,
                    minor=latest_version.minor,
                )
            )
            self.session.commit()
        elif current_version < latest_version:
            if auto_migrate:
                self.migrate()
            else:
                if not ignore_version:
                    raise DBVersionError(
                        f"Database version {current_version} needs migration but auto_migrate is disabled Please use snkmt db migrate command."
                    )

    def migrate(
        self,
        desired_version: Optional[DBVersion] = None,
        upgrade_only: bool = False,
        create_backup: bool = True,
    ) -> None:
        """
        Migrate database to desired version.

        Parameters
        ----------
        desired_version: Optional[DBVersionInfo]
            Desired version to update redun database to. If null, update to latest version.
        upgrade_only: bool
            By default, this function will perform both upgrades and downgrades.
            Set this to true to prevent downgrades (such as during automigration).
        create_backup: bool
            Create a timestamped backup of the database before migration.
        """
        assert self.engine
        assert self.session
        _, newest_allowed_version = self.get_db_version_required()

        if desired_version is None:
            desired_version = newest_allowed_version

        version = self.get_version()

        if version == desired_version:
            logger.info(
                f"Already at desired db version {version}. No migrations performed."
            )
            return

        if version > newest_allowed_version:
            # db version is too new to work with, abort.
            raise DBVersionError(
                f"Database is too new for this program: {version} > {newest_allowed_version}"
            )

        # Create backup before migration if requested and not a new database
        if create_backup and version != null_db_version:
            backup_path = self._create_backup()
            logger.info(f"Created database backup: {backup_path}")

        db_dir = Path(__file__).parent
        alembic_config_file = db_dir / "alembic.ini"
        alembic_script_location = db_dir / "alembic"

        config = AlembicConfig(alembic_config_file)
        config.set_main_option("script_location", str(alembic_script_location))
        config.session = self.session  # type: ignore

        # Perform migration.
        if desired_version > version:
            logger.info(f"Upgrading db from version {version} to {desired_version}...")
            upgrade(config, desired_version.id)
        elif desired_version < version and not upgrade_only:
            logger.info(
                f"Downgrading db from version {version} to {desired_version}..."
            )
            downgrade(config, desired_version.id)
        else:
            # Already at desired version.
            logger.info(f"Already at desired db version {version}.")
            return

        # Record migration has been applied.
        self.session.add(
            DBVersion(
                id=desired_version.id,
                major=desired_version.major,
                minor=desired_version.minor,
            )
        )

        # Commit migrations. This also ensures there are no checkedout connections.
        self.session.commit()

    def _create_backup(self) -> str:
        """Create a timestamped backup of the database file."""
        db_path = Path(self.db_path)
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        current_version_id = self.get_version().id
        backup_name = (
            f"{db_path.stem}_backup_{timestamp}_{current_version_id}_{db_path.suffix}"
        )
        backup_path = db_path.parent / backup_name

        # Close session temporarily to ensure file is not locked
        self.session.close()

        try:
            shutil.copy2(self.db_path, backup_path)
        finally:
            # Reopen session
            self.session = self.get_session()

        return str(backup_path)

    def get_version(self) -> DBVersion:
        """Get the current database version."""
        assert self.session
        inspector = inspect(self.engine)
        table_names = inspector.get_table_names()

        if DBVersion.__tablename__ in table_names:
            version_row = (
                self.session.query(DBVersion).order_by(DBVersion.major.desc()).first()
            )
            if version_row:
                return version_row

        return null_db_version

    @staticmethod
    def get_all_versions() -> List[DBVersion]:
        return DB_VERSIONS

    @staticmethod
    def get_db_version_required() -> Tuple[DBVersion, DBVersion]:
        """
        Returns the DB version range required by this library.
        """
        return DB_MIN_VERSION, DB_MAX_VERSION

    def get_session(self) -> Session:
        """New SQLAlchemy session."""
        return self.SessionLocal()

    def get_db_info(self) -> dict:
        """Path, tables, and engine URL."""
        inspector = inspect(self.engine)
        return {
            "db_path": self.db_path,
            "tables": inspector.get_table_names(),
            "engine": str(self.engine.url),
            "schema_revision": self.get_version().id,
        }



from pathlib import Path
from typing import List
from uuid import UUID
from sqlalchemy.orm import Session
from sqlalchemy import select
from textual.widgets import DataTable
from datetime import datetime, timezone
from rich.text import TextType, Text

from snkmt.db.models.rule import Rule
from snkmt.db.models.workflow import Workflow
from snkmt.db.models.enums import Status


class StyledProgress(Text):
    def __init__(self, progress: float) -> None:
        progstr = format(progress, ".2%")

        if progress < 0.2:
            color = "#fb4b4b"
        elif progress < 0.4:
            color = "#ffa879"
        elif progress < 0.6:
            color = "#ffc163"
        elif progress < 0.8:
            color = "#feff5c"
        else:
            color = "#c0ff33"
        super().__init__(progstr, style=color)


class StyledStatus(Text):
    def __init__(self, status: Status) -> None:
        status_str = status.value.capitalize()
        if status == Status.RUNNING:
            color = "#ffc163"
        elif status == Status.SUCCESS:
            color = "#c0ff33"
        elif status == Status.ERROR:
            color = "#fb4b4b"
        else:
            color = "#b0b0b0"
        super().__init__(status_str, style=color)


class RuleTable(DataTable):
    BINDINGS = [
        ("enter", "select_cursor", "Select"),
    ]

    def __init__(self, workflow_id: UUID, db_session: Session, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.workflow_id = workflow_id
        self.db_session = db_session
        self.last_update = None
        self._column_keys = self.add_columns(
            "Rule",
            "Progress",
            "# Jobs",
            "# Jobs Finished",
            "# Jobs Running",
            "# Jobs Pending",
            "# Jobs Failed",
        )
        self.cursor_type = "row"
        self.cursor_foreground_priority = "renderable"

    def on_mount(self) -> None:
        if self.workflow_id is None:
            return
        if self.last_update is None:
            # TODO move this query to Rule and setup limits/offsets like workflow
            rules = self.db_session.scalars(
                select(Rule).where(Rule.workflow_id == self.workflow_id)
            ).all()

            for rule in rules:
                row = self._rule_to_row(rule)
                self.add_row(
                    *row, key=rule.name
                )  # this could break if duplicate rule names. but i dont think that possible
            self.last_update = datetime.now(timezone.utc)
        self.set_interval(1, self._update)

    def _update(self) -> None:
        rules = Rule.get_updated_since(
            self.db_session, self.workflow_id, self.last_update
        )
        self.log.debug(f"Found {len(rules)} rules to update")
        for rule in rules:
            row = self._rule_to_row(rule)
            self._update_row(key=rule.name, row_data=row)
        self.last_update = datetime.now(timezone.utc)

    def _rule_to_row(self, rule: Rule) -> List[TextType]:
        job_counts = rule.get_job_counts(self.db_session)
        return [
            rule.name,
            StyledProgress(rule.progress),
            job_counts["total"],
            job_counts["success"],
            job_counts["running"],
            job_counts["pending"],
            job_counts["failed"],
        ]

    def _update_row(self, key: str, row_data: List[TextType]) -> None:
        """Update a single row, adding it if it doesn't exist."""
        if key not in self.rows:
            self.add_row(*row_data, key=key)
        else:
            existing_row = self.get_row(key)
            if existing_row != row_data:
                for col_idx, (new_val, old_val) in enumerate(
                    zip(row_data, existing_row)
                ):
                    if new_val != old_val:
                        column_key = self._column_keys[col_idx]
                        self.update_cell(key, column_key, new_val)


class WorkflowTable(DataTable):
    BINDINGS = [
        ("enter", "select_cursor", "Select"),
    ]

    def __init__(self, db_session: Session, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.db_session = db_session
        self.last_update = None
        self._column_keys = self.add_columns(
            "UUID", "Status", "Snakefile", "Started At", "Progress"
        )
        self.cursor_type = "row"
        self.cursor_foreground_priority = "renderable"

    def on_mount(self) -> None:
        if self.last_update is None:
            workflows = Workflow.list_all(self.db_session, limit=100)
            self.log.debug(f"Initial workflow table load: {len(workflows)} workflows.")

            for workflow in workflows:
                workflow_id = str(workflow.id)
                row_data = self._workflow_to_row(workflow)
                self.add_row(*row_data, key=workflow_id)
            self.last_update = datetime.now(timezone.utc)
        self.set_interval(0.5, self._update)

    def _update(self) -> None:
        workflows = Workflow.get_updated_since(self.db_session, self.last_update)
        self.log.debug(f"Found {len(workflows)} workflows to update")
        for workflow in workflows:
            self.log.debug(
                f"Updating workflow: {workflow.id}. {self.last_update=}, {workflow.updated_at=}"
            )
            row = self._workflow_to_row(workflow)
            self._update_row(key=str(workflow.id), row_data=row)
        self.last_update = datetime.now(timezone.utc)

    def _workflow_to_row(self, workflow: Workflow) -> List[TextType]:
        workflow_id = str(workflow.id)
        status = StyledStatus(workflow.status)
        snakefile = Path(workflow.snakefile).name if workflow.snakefile else "N/A"
        started_at = (
            workflow.started_at.strftime("%Y-%m-%d %H:%M:%S")
            if workflow.started_at
            else "N/A"
        )
        progress = StyledProgress(workflow.progress)
        return [workflow_id[-6:], status, snakefile, started_at, progress]

    def _update_row(self, key: str, row_data: List[TextType]) -> None:
        """Update a single row, adding it if it doesn't exist."""
        if key not in self.rows:
            self.add_row(*row_data, key=key)
        else:
            existing_row = self.get_row(key)
            if existing_row != row_data:
                for col_idx, (new_val, old_val) in enumerate(
                    zip(row_data, existing_row)
                ):
                    if new_val != old_val:
                        column_key = self._column_keys[col_idx]
                        self.update_cell(key, column_key, new_val)



import typer
from typing import Optional
from snkmt.db.session import Database

app = typer.Typer(
    name="snkmt",
    help="Monitor Snakemake workflow executions.",
    add_completion=False,
    no_args_is_help=True,
)

db_app = typer.Typer()
app.add_typer(db_app, name="db")


### MAIN APP COMMANDS
@app.callback()
def callback():
    pass


@app.command("console")
def launch_console(
    directory: Optional[str] = typer.Option(
        None, "--db-path", "-d", help="Path to the database."
    ),
):
    """Launch the interactive console UI"""
    from snkmt.console.app import run_app

    db = Database(db_path=directory).get_session()
    run_app(db)


#### DB APP COMMANDS
@db_app.callback()
def db_callback():
    pass


@db_app.command("info")
def db_info(db: Optional[str]):
    database = Database(db, create_db=False, auto_migrate=False)
    print(f"Database info: {database.get_db_info()}")


@db_app.command("migrate")
def db_migrate(db: Optional[str]):
    database = Database(db, create_db=False, auto_migrate=False, ignore_version=True)
    database.migrate()


def main():
    app()



from enum import Enum


class Status(Enum):
    RUNNING = "RUNNING"
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"
    UNKNOWN = "UNKNOWN"


class FileType(Enum):
    INPUT = "INPUT"
    OUTPUT = "OUTPUT"
    LOG = "LOG"
    BENCHMARK = "BENCHMARK"



from typing import Any
from snkmt.db.models.base import Base
from sqlalchemy.orm import Mapped, mapped_column
from sqlalchemy import String
from datetime import datetime, timezone

# db versioning adapted from https://github.com/insitro/redun/blob/main/redun/backends/db/__init__.py
DB_UNKNOWN_VERSION = 99


class DBVersionError(Exception):
    pass


class DBVersion(Base):
    """
    Database version
    """

    __tablename__ = "snkmt_db_version"
    id: Mapped[str] = mapped_column(
        String, primary_key=True
    )  # this can just be the alembic revision id
    major: Mapped[int] = mapped_column()
    minor: Mapped[int] = mapped_column()
    timestamp: Mapped[datetime] = mapped_column(
        default=lambda: datetime.now(timezone.utc),
        onupdate=lambda: datetime.now(timezone.utc),
    )

    def __lt__(self, other) -> bool:
        if not isinstance(other, DBVersion):
            raise TypeError(f"Expected DBVersion: {other}")
        return (self.major, self.minor) < (other.major, other.minor)

    def __le__(self, other) -> bool:
        if not isinstance(other, DBVersion):
            raise TypeError(f"Expected DBVersion: {other}")
        return (self.major, self.minor) <= (other.major, other.minor)

    def __gt__(self, other) -> bool:
        if not isinstance(other, DBVersion):
            raise TypeError(f"Expected DBVersion: {other}")
        return (self.major, self.minor) > (other.major, other.minor)

    def __ge__(self, other) -> bool:
        if not isinstance(other, DBVersion):
            raise TypeError(f"Expected DBVersion: {other}")
        return (self.major, self.minor) >= (other.major, other.minor)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, DBVersion):
            raise TypeError(f"Expected DBVersion: {other}")
        return (self.major, self.minor) == (other.major, other.minor)

    def __str__(self) -> str:
        if self.minor == DB_UNKNOWN_VERSION:
            return f"{self.major}.?"
        else:
            return f"{self.major}.{self.minor}"


# List of all available database versions and migrations.
# Note these are sorted from oldest to newest.
null_db_version = DBVersion(id="000000000000", major=-1, minor=0)
DB_VERSIONS = [DBVersion(id="a088a7b93fe5", major=1, minor=0)]
DB_MIN_VERSION = DBVersion(id="", major=1, minor=0)  # Min db version needed by snkmt.
DB_MAX_VERSION = DBVersion(id="", major=1, minor=0)  # Max db version needed by snkmt.


def parse_db_version(version_str: str) -> DBVersion:
    """
    Parses a db version string such as "2.0" or "3" into a DBVersion.
    """
    if version_str == "latest":
        return DB_VERSIONS[-1]

    dots = version_str.count(".")
    if dots == 0:
        major, minor = (int(version_str), 0)
    elif dots == 1:
        major, minor = tuple(map(int, version_str.split(".")))
    else:
        raise ValueError(f"Invalid db version format: {version_str}")

    for version in DB_VERSIONS:
        if version.major == major and version.minor == minor:
            return version

    raise DBVersionError(f"Unknown db version: {version_str}")



from typing import TYPE_CHECKING

from sqlalchemy import ForeignKey, Enum
from sqlalchemy.orm import Mapped, mapped_column, relationship

from snkmt.db.models.base import Base
from snkmt.db.models.enums import FileType

if TYPE_CHECKING:
    from snkmt.db.models.job import Job


class File(Base):
    __tablename__ = "files"
    id: Mapped[int] = mapped_column(primary_key=True)
    path: Mapped[str]  # TODO: use pathlib.Path/os.pathlike type here eventually
    file_type: Mapped[FileType] = mapped_column(Enum(FileType))
    job_id: Mapped[int] = mapped_column(ForeignKey("jobs.id"))
    job: Mapped["Job"] = relationship("Job", back_populates="files")



import pytest
from pathlib import Path
from snkmt.db.session import Database
from snkmt.db.models.version import DBVersion
import tempfile


@pytest.fixture
def temp_db_path():
    """Create a temporary database path."""
    with tempfile.TemporaryDirectory() as temp_dir:
        yield Path(temp_dir) / "test.db"


def test_new_database_sets_latest_version(temp_db_path):
    """Test that a new database is set to the latest version."""

    db = Database(db_path=str(temp_db_path), create_db=True)

    actual_version = db.get_version()
    expected_version = DBVersion(id="a088a7b93fe5", major=1, minor=0)

    assert actual_version == expected_version


def test_db_model():
    """Test DBVersion model comparison operators and string representation."""
    v1_0 = DBVersion(id="test1", major=1, minor=0)
    v1_1 = DBVersion(id="test2", major=1, minor=1)
    v2_0 = DBVersion(id="test3", major=2, minor=0)
    v1_0_duplicate = DBVersion(id="test4", major=1, minor=0)
    v_unknown = DBVersion(id="test5", major=1, minor=99)  # DB_UNKNOWN_VERSION

    # Test equality
    assert v1_0 == v1_0_duplicate
    assert not (v1_0 == v1_1)

    # Test less than
    assert v1_0 < v1_1
    assert v1_1 < v2_0
    assert not (v1_1 < v1_0)

    # Test less than or equal
    assert v1_0 <= v1_0_duplicate
    assert v1_0 <= v1_1
    assert not (v1_1 <= v1_0)

    # Test greater than
    assert v1_1 > v1_0
    assert v2_0 > v1_1
    assert not (v1_0 > v1_1)

    # Test greater than or equal
    assert v1_0 >= v1_0_duplicate
    assert v1_1 >= v1_0
    assert not (v1_0 >= v1_1)

    # Test string representation
    assert str(v1_0) == "1.0"
    assert str(v1_1) == "1.1"
    assert str(v2_0) == "2.0"
    assert str(v_unknown) == "1.?"

    # Test TypeError for invalid comparisons
    with pytest.raises(TypeError):
        v1_0 < "not_a_version"

    with pytest.raises(TypeError):
        v1_0 == "not_a_version"
